from optparse import make_option

from django.core.management.base import BaseCommand

from restraint.core import update_restraint_db


class Command(BaseCommand):
    """
    A management command for updating the restraint permissions based on the
    restraint config.
    """
    option_list = BaseCommand.option_list + (
        make_option(
            '--flush_default_access',
            action='store_true',
            dest='flush_default_access',
            default=False,
            help='Flush all permission sets before updating'
        ),
    )

    def handle(self, *args, **options):
        """
        Runs the command to update the restraint db.
        """
        update_restraint_db(flush_default_access=options['flush_default_access'])



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'PermSet'
        db.create_table(u'restraint_permset', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=256, blank=True)),
            ('display_name', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'restraint', ['PermSet'])

        # Adding model 'Perm'
        db.create_table(u'restraint_perm', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=256, blank=True)),
            ('display_name', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'restraint', ['Perm'])

        # Adding model 'PermLevel'
        db.create_table(u'restraint_permlevel', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('perm', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['restraint.Perm'])),
            ('name', self.gf('django.db.models.fields.CharField')(max_length=256, blank=True)),
            ('display_name', self.gf('django.db.models.fields.TextField')(blank=True)),
        ))
        db.send_create_signal(u'restraint', ['PermLevel'])

        # Adding unique constraint on 'PermLevel', fields ['perm', 'name']
        db.create_unique(u'restraint_permlevel', ['perm_id', 'name'])

        # Adding model 'PermAccess'
        db.create_table(u'restraint_permaccess', (
            (u'id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('perm_set', self.gf('django.db.models.fields.related.OneToOneField')(to=orm['restraint.PermSet'], unique=True)),
        ))
        db.send_create_signal(u'restraint', ['PermAccess'])

        # Adding M2M table for field perm_levels on 'PermAccess'
        m2m_table_name = db.shorten_name(u'restraint_permaccess_perm_levels')
        db.create_table(m2m_table_name, (
            ('id', models.AutoField(verbose_name='ID', primary_key=True, auto_created=True)),
            ('permaccess', models.ForeignKey(orm[u'restraint.permaccess'], null=False)),
            ('permlevel', models.ForeignKey(orm[u'restraint.permlevel'], null=False))
        ))
        db.create_unique(m2m_table_name, ['permaccess_id', 'permlevel_id'])


    def backwards(self, orm):
        # Removing unique constraint on 'PermLevel', fields ['perm', 'name']
        db.delete_unique(u'restraint_permlevel', ['perm_id', 'name'])

        # Deleting model 'PermSet'
        db.delete_table(u'restraint_permset')

        # Deleting model 'Perm'
        db.delete_table(u'restraint_perm')

        # Deleting model 'PermLevel'
        db.delete_table(u'restraint_permlevel')

        # Deleting model 'PermAccess'
        db.delete_table(u'restraint_permaccess')

        # Removing M2M table for field perm_levels on 'PermAccess'
        db.delete_table(db.shorten_name(u'restraint_permaccess_perm_levels'))


    models = {
        u'restraint.perm': {
            'Meta': {'object_name': 'Perm'},
            'display_name': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '256', 'blank': 'True'})
        },
        u'restraint.permaccess': {
            'Meta': {'object_name': 'PermAccess'},
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'perm_levels': ('django.db.models.fields.related.ManyToManyField', [], {'to': u"orm['restraint.PermLevel']", 'symmetrical': 'False'}),
            'perm_set': ('django.db.models.fields.related.OneToOneField', [], {'to': u"orm['restraint.PermSet']", 'unique': 'True'})
        },
        u'restraint.permlevel': {
            'Meta': {'unique_together': "(('perm', 'name'),)", 'object_name': 'PermLevel'},
            'display_name': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '256', 'blank': 'True'}),
            'perm': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['restraint.Perm']"})
        },
        u'restraint.permset': {
            'Meta': {'object_name': 'PermSet'},
            'display_name': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '256', 'blank': 'True'})
        }
    }

    complete_apps = ['restraint']


from django.db import models
from manager_utils import sync


class PermSetManager(models.Manager):
    def sync_perm_sets(self, perm_sets):
        """
        Syncs the provided dictionary of perm sets to PermSet models.
        """
        sync(
            self.get_queryset(),
            [PermSet(name=name, display_name=config.get('display_name', '')) for name, config in perm_sets.items()],
            ['name'], ['display_name'])


class PermSet(models.Model):
    name = models.CharField(max_length=256, unique=True, blank=True)
    display_name = models.TextField(blank=True)

    objects = PermSetManager()

    def __unicode__(self):
        return self.display_name


class PermManager(models.Manager):
    def sync_perms(self, perms):
        """
        Syncs the perms to Perm models.
        """
        sync(
            self.get_queryset(),
            [Perm(name=name, display_name=config.get('display_name', '')) for name, config in perms.items()],
            ['name'], ['display_name'])


class Perm(models.Model):
    name = models.CharField(max_length=256, unique=True, blank=True)
    display_name = models.TextField(blank=True)

    objects = PermManager()

    def __unicode__(self):
        return self.display_name


class PermLevelManager(models.Manager):
    def sync_perm_levels(self, perms):
        """
        Given a dictionary of perms that map to perm levels, sync the perm levels
        to PermLevel objects in the database.
        """
        perm_objs = {p.name: p for p in Perm.objects.all()}
        perm_levels = []
        for perm, perm_config in perms.items():
            assert(perm_config['levels'])
            for l, level_config in perm_config['levels'].items():
                perm_levels.append(
                    PermLevel(perm=perm_objs[perm], name=l, display_name=level_config.get('display_name', '')))

        sync(self.get_queryset(), perm_levels, ['name', 'perm'], ['display_name'])


class PermLevel(models.Model):
    perm = models.ForeignKey(Perm)
    name = models.CharField(max_length=256, blank=True)
    display_name = models.TextField(blank=True)

    objects = PermLevelManager()

    class Meta:
        unique_together = ('perm', 'name')

    def __unicode__(self):
        return self.display_name


class PermAccessManager(models.Manager):
    def update_perm_set_access(self, config, flush_previous_config=False):
        """
        Update the access for perm sets with a config. The user can optionally flush
        the previous config and set it to the new one.
        """
        for perm_set in PermSet.objects.all():
            perm_access, created = PermAccess.objects.get_or_create(perm_set=perm_set)
            if not created and not flush_previous_config:
                # If we are not flushing the previous config, continue if it exists
                continue

            perm_access_levels = []
            for perm, perm_levels in config.get(perm_set.name, {}).items():
                assert(perm_levels)
                perm_access_levels.extend(PermLevel.objects.filter(perm__name=perm, name__in=perm_levels))

            perm_access.perm_levels.clear()
            perm_access.perm_levels.add(*perm_access_levels)


class PermAccess(models.Model):
    perm_set = models.OneToOneField(PermSet)
    perm_levels = models.ManyToManyField(PermLevel)

    objects = PermAccessManager()



from django.core.management import call_command
from django.test import SimpleTestCase
from mock import patch


class UpdateRestraintDbTest(SimpleTestCase):
    @patch('restraint.management.commands.update_restraint_db.update_restraint_db', spec_set=True)
    def test_wo_flush_default_access(self, mock_update_restraint_db):
        call_command('update_restraint_db')
        mock_update_restraint_db.assert_called_once_with(flush_default_access=False)

    @patch('restraint.management.commands.update_restraint_db.update_restraint_db', spec_set=True)
    def test_w_flush_default_access(self, mock_update_restraint_db):
        call_command('update_restraint_db', flush_default_access=True)
        mock_update_restraint_db.assert_called_once_with(flush_default_access=True)



from django.test import SimpleTestCase
from django_dynamic_fixture import N

from restraint.models import PermSet, Perm, PermLevel


class PermSetTest(SimpleTestCase):
    def test_unicode(self):
        ps = N(PermSet, display_name='My Perm Set', persist_dependencies=False)
        self.assertTrue(ps.__unicode__(), 'My Perm Set')


class PermTest(SimpleTestCase):
    def test_unicode(self):
        ps = N(Perm, display_name='My Perm', persist_dependencies=False)
        self.assertTrue(ps.__unicode__(), 'My Perm')


class PermLevelTest(SimpleTestCase):
    def test_unicode(self):
        ps = N(PermLevel, display_name='My Perm Level', persist_dependencies=False)
        self.assertTrue(ps.__unicode__(), 'My Perm Level')



from collections import defaultdict
from itertools import chain

from restraint import models


# A global variable for holding the configuration of django restraint
RESTRAINT_CONFIG = {}


def register_restraint_config(restraint_config):
    RESTRAINT_CONFIG.clear()
    RESTRAINT_CONFIG.update(restraint_config)


def get_restraint_config():
    if RESTRAINT_CONFIG:
        return RESTRAINT_CONFIG
    else:
        raise RuntimeError('No restraint config has been registered')


def update_restraint_db(flush_default_access=False):
    """
    Updates the restraint db based on the restraint config.
    Can optionally flush the previous default access configuration.
    """
    config = get_restraint_config()
    models.PermSet.objects.sync_perm_sets(config['perm_sets'])
    models.Perm.objects.sync_perms(config['perms'])
    models.PermLevel.objects.sync_perm_levels(config['perms'])
    models.PermAccess.objects.update_perm_set_access(config.get('default_access', {}), flush_default_access)


class Restraint(object):
    """
    The primary way of accessing permissions. The programmer loads a restraint object with the
    permission object and which permissions they want to load. One permissions are loaded for
    that account, the user may check if a user has certain permissions and also restrict
    querysets based on access levels that a user has.
    """
    def __init__(self, user, which_perms=None):
        """
        Initializes the Restraint object.

        :type user: Any object
        :param user: A user in a project

        :type which_perms: list
        :param which_perms: The permissions to be loaded for the user, or all permissions if None.
        """
        self._config = get_restraint_config()
        self._user = user
        self._load_perms(user, which_perms)

    @property
    def perms(self):
        return self._perms

    def _load_perms(self, account, which_perms):
        perm_set_names = self._config['perm_set_getter'](account)
        perm_levels = models.PermLevel.objects.filter(
            permaccess__perm_set__name__in=perm_set_names).select_related('perm')
        if which_perms:
            perm_levels = perm_levels.filter(perm__name__in=which_perms)

        self._perms = defaultdict(dict)
        for l in perm_levels:
            self._perms[l.perm.name].update({
                l.name: self._config['perms'][l.perm.name]['levels'][l.name]['id_filter']
            })

    def has_perm(self, perm, level=None):
        """
        Returns true if the restraint object has the perm. If a level is not specified, it returns
        true if that perm exists for any level.

        :type perm: string
        :param perm: The permission to check

        :type level: string
        :param level: The level to check, or any level if None
        """
        return perm in self._perms and level in self._perms[perm] if level else perm in self._perms

    def filter_qset(self, qset, perm):
        """
        Given a permission, filter the queryset by its levels.

        :type qset: A Django QuerySet
        :param qset: The queryset to be filtered

        :type perm: string
        :param perm: The permission over which to do the filtering
        """
        if not self.has_perm(perm):
            # If the user doesnt have the perm, return no data
            return qset.none()
        elif None in self._perms[perm].values():
            # If any levels are none, return the full queryset
            return qset
        else:
            # Filter the queryset by the union of all filters
            return qset.filter(id__in=set(chain(*[l(self._user) for l in self._perms[perm].values()])))



from django.contrib.auth.models import User
from django.test import SimpleTestCase, TestCase
from django_dynamic_fixture import G
from mock import patch, Mock

from restraint import core, constants
from restraint.models import PermSet, Perm, PermLevel, PermAccess


class TestRegisterRestraintConfig(SimpleTestCase):
    def test_register_restraint_config(self):
        core.register_restraint_config({'config': 'config'})
        self.assertEquals(core.RESTRAINT_CONFIG, {'config': 'config'})


class TestGetRestraintConfig(SimpleTestCase):
    def test_get_restraint_config_not_set(self):
        core.RESTRAINT_CONFIG.clear()
        with self.assertRaises(RuntimeError):
            core.get_restraint_config()

    def test_get_restraint_config_set(self):
        core.register_restraint_config({'config': 'config'})
        self.assertEquals(core.get_restraint_config(), {'config': 'config'})


class TestRestraintLoadPerms(TestCase):
    def setUp(self):
        def perm_set_getter(u):
            perm_sets = ['individual']
            if u.is_superuser:
                perm_sets.append('super')
            return perm_sets

        config = {
            'perm_set_getter': perm_set_getter,
            'perm_sets': {
                'super': {
                    'display_name': 'Super',
                },
                'individual': {
                    'display_name': 'Individual'
                }
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        }
                    }
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                }
            },
            'default_access': {
                'super': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'individual': {
                    'can_edit_stuff': ['some_stuff'],
                }
            }
        }
        core.register_restraint_config(config)
        core.update_restraint_db()

    def test_load_all_perms(self):
        # Make a user that is a superuser and verify they get all of the
        # super user perms
        u = G(User, is_superuser=True)
        r = core.Restraint(u)
        perms = r.perms
        self.assertEquals(perms, {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        })

    def test_load_some_perms(self):
        # Make a user that is a superuser and verify they get all of the
        # super user perms
        u = G(User, is_superuser=True)
        r = core.Restraint(u, ['can_edit_stuff'])
        perms = r.perms
        self.assertEquals(perms, {
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        })


class TestRestraintHasPerms(SimpleTestCase):
    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_w_level_true(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertTrue(r.has_perm('can_edit_stuff', 'all_stuff'))

    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_w_level_false(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertFalse(r.has_perm('can_edit_stuff', 'no_stuff'))

    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_wo_level_true(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertTrue(r.has_perm('can_edit_stuff'))

    @patch.object(core.Restraint, '_load_perms', spec_set=True)
    def test_has_perm_wo_level_false(self, mock_load_perms):
        r = core.Restraint(Mock())
        r._perms = {
            'can_view_stuff': {
                '': None,
            },
            'can_edit_stuff': {
                'all_stuff': None,
                'some_stuff': None,
            }
        }
        self.assertFalse(r.has_perm('can_mess_with_stuff'))


class TestRestraintFilterQSet(TestCase):
    def setUp(self):
        def perm_set_getter(u):
            perm_sets = ['individual']
            if u.is_superuser:
                perm_sets.append('super')
            if u.is_staff:
                perm_sets.append('staff')
            return perm_sets

        config = {
            'perm_set_getter': perm_set_getter,
            'perm_sets': {
                'super': {
                    'display_name': 'Super',
                },
                'individual': {
                    'display_name': 'Individual',
                },
                'staff': {
                    'display_name': 'Staff'
                }
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': lambda a: User.objects.filter(id=a.id).values_list('id', flat=True),
                        },
                        'only_superusers': {
                            'display_name': 'Only Superusers',
                            'id_filter': lambda a: User.objects.filter(is_superuser=True).values_list('id', flat=True),
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                }
            },
            'default_access': {
                'super': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'individual': {
                    'can_edit_stuff': ['some_stuff'],
                },
                'staff': {
                    'can_edit_stuff': ['some_stuff', 'only_superusers']
                }
            }
        }
        core.register_restraint_config(config)
        core.update_restraint_db()

    def test_filter_qset_global_access(self):
        # Make a user that is a superuser and verify they get all of the
        # super user perms
        u = G(User, is_superuser=True)
        # Make another user that they will be able to edit
        u2 = G(User)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([u, u2]))

    def test_filter_qset_local_access(self):
        # Make a user that is not a superuser
        u = G(User, is_superuser=False)
        # Make another user that they will not be able to edit
        G(User)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([u]))

    def test_filter_qset_multiple_local_access(self):
        # Make a user that is staff
        u = G(User, is_superuser=False, is_staff=True)
        # Make another user that they will not be able to edit
        G(User)
        # Make another super user that they will be able to edit
        u2 = G(User, is_superuser=True)
        r = core.Restraint(u)

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([u, u2]))

    def test_filter_qset_no_perms(self):
        # Make a user that is staff
        u = G(User, is_superuser=False, is_staff=True)
        # Load permissions that will not give them access to edit any accounts
        r = core.Restraint(u, ['bad_perm'])

        filtered_qset = r.filter_qset(User.objects.all(), 'can_edit_stuff')
        self.assertEquals(set(filtered_qset), set([]))


class UpdateRestraintDbTest(TestCase):
    def test_full_update_scenario_not_flush_default_access(self):
        core.register_restraint_config({
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'restricted': {
                    'can_edit_stuff': ['some_stuff'],
                }
            }
        })
        core.update_restraint_db()
        core.update_restraint_db()

        self.assertEquals(
            set(PermSet.objects.values_list('name', flat=True)), set(['global', 'restricted']))

        self.assertEquals(
            set(Perm.objects.values_list('name', flat=True)), set(['can_view_stuff', 'can_edit_stuff']))

        self.assertEquals(
            set(PermLevel.objects.values_list('name', 'perm__name')),
            set([('all_stuff', 'can_edit_stuff'), ('some_stuff', 'can_edit_stuff'), ('', 'can_view_stuff')]))

        self.assertEquals(
            set(PermAccess.objects.values_list('perm_levels__name', 'perm_levels__perm__name', 'perm_set__name')),
            set([
                ('all_stuff', 'can_edit_stuff', 'global'), ('', 'can_view_stuff', 'global'),
                ('some_stuff', 'can_edit_stuff', 'global'), ('some_stuff', 'can_edit_stuff', 'restricted')
            ])
        )

    def test_full_update_scenario_flush_default_access(self):
        core.register_restraint_config({
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff', 'some_stuff'],
                    'can_view_stuff': [constants.BOOLEAN_LEVELS_NAME],
                },
                'restricted': {
                    'can_edit_stuff': ['some_stuff'],
                }
            }
        })
        core.update_restraint_db()

        core.register_restraint_config({
            'perm_sets': {
                'global': {
                    'display_name': 'Global',
                },
                'restricted': {
                    'display_name': 'Restricted',
                },
            },
            'perms': {
                'can_edit_stuff': {
                    'display_name': 'Can Edit Stuff',
                    'levels': {
                        'all_stuff': {
                            'display_name': 'All Stuff',
                            'id_filter': None,
                        },
                        'some_stuff': {
                            'display_name': 'Some Stuff',
                            'id_filter': None,
                        },
                    },
                },
                'can_view_stuff': {
                    'display_name': 'Can View Stuff',
                    'levels': constants.BOOLEAN_LEVELS_CONFIG,
                },
            },
            'default_access': {
                'global': {
                    'can_edit_stuff': ['all_stuff'],
                },
            }
        })
        core.update_restraint_db(flush_default_access=True)

        self.assertEquals(
            set(PermSet.objects.values_list('name', flat=True)), set(['global', 'restricted']))

        self.assertEquals(
            set(Perm.objects.values_list('name', flat=True)), set(['can_view_stuff', 'can_edit_stuff']))

        self.assertEquals(
            set(PermLevel.objects.values_list('name', 'perm__name')),
            set([('all_stuff', 'can_edit_stuff'), ('some_stuff', 'can_edit_stuff'), ('', 'can_view_stuff')]))

        self.assertEquals(
            set(PermAccess.objects.values_list('perm_levels__name', 'perm_levels__perm__name', 'perm_set__name')),
            set([('all_stuff', 'can_edit_stuff', 'global'), (None, None, 'restricted')])
        )
